"""Global configuration options."""

from dataclasses import dataclass


@dataclass
class ConfigOptions:
    offline_mode: bool = False
    _online_source: str = "cdn"
    _offline_source: str = "inline"

    matplotlib_output_format: str = "svg"
    matplotlib_notebook_format: str = "png"

    pdf_temp_dir: str = ".pdf-temp"


options = ConfigOptions()


def get_source_from_options(source):
    if source == "esparto.options":
        if options.offline_mode:
            return options._offline_source
        else:
            return options._online_source
    elif source in ["cdn", "inline"]:
        return source
    raise ValueError(f"Unrecognised source: {source}")



import re
from typing import Callable, Iterable, List


def get_index_where(condition: Callable[..., bool], iterable: Iterable) -> List[int]:
    """Return index values where `condition` is `True`."""
    return [idx for idx, item in enumerate(iterable) if condition(item)]


def get_matching_titles(title: str, children: list) -> List[int]:
    """Return child items with matching title."""
    return get_index_where(lambda x: getattr(x, "title", None) == title, children)


def clean_attr_name(attr_name: str):
    # Remove leading and trailing spaces
    attr_name = attr_name.strip().replace(" ", "_").lower()

    # Remove invalid characters
    attr_name = re.sub("[^0-9a-zA-Z_]", "", attr_name)

    # Remove leading characters until we find a letter or underscore
    attr_name = re.sub("^[^a-zA-Z_]+", "", attr_name)

    return attr_name


def clean_iterator(iterator: Iterable) -> Iterable:
    # Convert any non-list iterators to lists
    iterator = (
        list(iterator) if isinstance(iterator, (list, tuple, set)) else [iterator]
    )
    # Unnest any nested lists of children
    if len(list(iterator)) == 1 and isinstance(list(iterator)[0], (list, tuple, set)):
        iterator = list(iterator)[0]
    return iterator


def responsive_svg_mpl(source: str, width: int = None, height: int = None) -> str:
    """Make SVG element responsive."""

    regex_w = r"width=\S*"
    regex_h = r"height=\S*"

    width_ = f"width='{width}px'" if width else ""
    height_ = f"height='{height}px'" if height else ""

    source = re.sub(regex_w, width_, source, count=1)
    source = re.sub(regex_h, height_, source, count=1)

    # Preserve aspect ratio of SVG
    regexp = r"<svg"
    repl = '<svg class="svg-content-mpl" preserveAspectRatio="xMinYMin meet" '
    source = re.sub(regexp, repl, source, count=1)

    return source



"""Layout classes for defining and interacting with a document."""

import copy
from abc import ABC
from pprint import pformat
from typing import TYPE_CHECKING, Any, Dict, Iterable, List, Optional, Set, Type, Union

from esparto._publish import nb_display, publish_html, publish_pdf
from esparto._utils import clean_attr_name, clean_iterator, get_matching_titles

if TYPE_CHECKING:
    from esparto._content import Content


class Layout(ABC):
    """Template for Layout elements. All Layout classes come with these methods and attributes.

    Layout class hierarchy:
        `Page -> Section -> Row -> Column -> Content`

    Attributes:
      title (str): Object title. Used as a title within the document and as a key value.
      children (list): Child items defining the document layout and content.

    """

    # ------------------------------------------------------------------------+
    #                              Magic Methods                              |
    # ------------------------------------------------------------------------+

    def __init__(
        self,
        title: Optional[str] = None,
        children: Union[
            List[Union["Layout", "Content", Any]], "Layout", "Content"
        ] = list(),
    ):
        self.set_children(children)
        self.title = title

    def __iter__(self):
        return iter([self])

    def __repr__(self):
        return self._tree()

    def _repr_html_(self):
        self.display()

    def __str__(self):
        return self._tree()

    def __add__(self, other: Union["Layout", "Content", Any]):

        if isinstance(other, type(self)):
            return self._parent_class(
                title=self.title, children=[*(*self.children, *other.children)]
            )

        new = copy.copy(self)
        new.children = self.children + [*self._smart_wrap(other)]

        return new

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return (
                self.title == other.title
                and len(self.children) == len(other.children)
                and all((x == y for x, y in zip(self.children, other.children)))
            )
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __getattribute__(self, key: str) -> Any:
        child_id = super().__getattribute__("_child_ids").get(key)
        if child_id:
            return self.__getitem__(child_id)
        return super().__getattribute__(key)

    def __setattr__(self, key: str, value: Any) -> None:
        child_id = super().__getattribute__("_child_ids").get(key)
        if child_id:
            self.__setitem__(child_id, value)
        else:
            super().__setattr__(key, value)

    def __delattr__(self, key: str) -> None:
        child_id = super().__getattribute__("_child_ids").get(key)
        if child_id:
            self.__delitem__(child_id)
        else:
            super().__delattr__(key)

    def __getitem__(self, key: Union[str, int]):
        if isinstance(key, str):
            indexes = get_matching_titles(key, self.children)
            if len(indexes):
                return self.children[indexes[0]]
            value = self._child_class(title=key)
            self.children.append(value)
            if key:
                self._add_child_id(key)
            return self[key]

        elif isinstance(key, int) and key < len(self.children):
            return self.children[key]

        raise KeyError(key)

    def __setitem__(self, key: Union[str, int], value: Any):
        value = copy.copy(value)
        value = self._smart_wrap(value)
        value = value[0]
        if isinstance(key, str):
            if key:
                value.title = key
                indexes = get_matching_titles(key, self.children)
                if indexes:
                    self.children[indexes[0]] = value
                else:
                    self.children.append(value)
                self._add_child_id(key)
            else:
                self.children.append(value)
            return None
        elif isinstance(key, int) and key < len(self.children):
            value.title = getattr(self.children[key], "title", None)
            self.children[key] = value
            return None
        raise KeyError(key)

    def __delitem__(self, key) -> None:
        if isinstance(key, str):
            indexes = get_matching_titles(key, self.children)
            if len(indexes):
                self._remove_child_id(key)
                del self.children[indexes[0]]
                return None
        elif isinstance(key, int) and key < len(self.children):
            child_title = getattr(self.children[key], "title", None)
            self._remove_child_id(child_title)
            del self.children[key]
            return None
        raise KeyError(key)

    def __lshift__(self, other: Union["Layout", "Content", Any]):
        self.set_children(other)
        return other

    def __rshift__(self, other: Union["Layout", "Content", Any]):
        self.set_children(other)
        return self

    title: Optional[str]
    children: List[Any] = []
    _parent_class: Type["Layout"]
    _child_class: Type["Layout"]
    _title_tags: str
    _body_tags: str
    _dependencies = {"bootstrap"}

    @property
    def _child_ids(self) -> Dict[str, str]:
        """Return existing child IDs or a new dict."""
        try:
            super().__getattribute__("__child_ids")
        except AttributeError:
            super().__setattr__("__child_ids", dict())
        return super().__getattribute__("__child_ids")

    # ------------------------------------------------------------------------+
    #                              Public Methods                             |
    # ------------------------------------------------------------------------+

    def display(self) -> None:
        """Display rendered document in a Notebook environment."""
        nb_display(self)

    def set_children(self, other: Union["Layout", "Content", Any]):
        """Set children as other."""
        other = copy.copy(other)
        self.children = [*self._smart_wrap(other)]

    def to_html(self, **kwargs) -> str:
        """Convert document to HTML code.

        Returns:
          str: HTML code.

        """
        children_rendered = " ".join([c.to_html(**kwargs) for c in self.children])
        title_rendered = self._title_tags.format(self.title) if self.title else ""

        html = self._body_tags.format(f"{title_rendered}\n{children_rendered}\n")
        return html

    def tree(self) -> None:
        """Display document tree."""
        print(self._tree())

    # ------------------------------------------------------------------------+
    #                             Private Methods                             |
    # ------------------------------------------------------------------------+

    def _add_child_id(self, key):
        attr_name = clean_attr_name(key)
        if attr_name:
            self._child_ids[attr_name] = key
            super().__setattr__(attr_name, self[key])

    def _remove_child_id(self, key):
        attr_name = clean_attr_name(key)
        if attr_name in self._child_ids:
            del self._child_ids[attr_name]
            super().__delattr__(attr_name)

    def _smart_wrap(
        self, child_list: Iterable[Any]
    ) -> Iterable[Union["Layout", "Content"]]:
        """Wrap children in a coherent class hierarchy.

        Args:
          children: Sequence of Content and / or Child items.

        Returns:
          List of Layout and Content items wrapped in a coherent class hierarchy.


        If the parent object is a Column and the item is a Content Class:
            - return child with no modification
        If the parent object is a Column and the item is not a Content Class:
            - cast the child to an appropriate Content Class if possible
            - return the child
        If the current item is wrapped and unwrapped items have been accumulated:
            - wrap the unwrapped children
            - append newly wrapped to output
            - append current child to output
        If the current child is wrapped and we have no accumulated unwrapped items:
            - append the wrapped child to output
        If the current child is unwrapped and the parent is a Row:
            - wrap and append the current child to output
        If the current item is unwrapped and the parent is not a Row:
            - add the current child to unwrapped item accumulator
        Finally:
            - wrap any accumulated unwrapped items
            - append the final wrapped segment to output

        """
        from esparto._adaptors import content_adaptor

        child_list = clean_iterator(child_list)

        if isinstance(self, Column):
            return [content_adaptor(x) for x in child_list]

        is_row = isinstance(self, Row)
        unwrapped_acc: list = []
        output = []

        for child in child_list:
            is_wrapped = isinstance(child, self._child_class)

            if is_wrapped:
                if unwrapped_acc:
                    wrapped_segment = self._child_class(children=unwrapped_acc)
                    output.append(wrapped_segment)
                    output.append(child)
                    unwrapped_acc = []
                else:
                    output.append(child)
            else:  # if not is_wrapped
                if is_row:
                    output.append(self._child_class(children=[child]))
                else:
                    unwrapped_acc.append(child)

        if unwrapped_acc:
            wrapped_segment = self._child_class(children=unwrapped_acc)
            output.append(wrapped_segment)

        return output

    def _recurse_children(self, idx) -> dict:
        key = self.title or f"{type(self).__name__} {idx}"
        tree = {
            f"{key}": [
                child._recurse_children(idx)  # type: ignore
                if hasattr(child, "_recurse_children")
                else str(child)
                for idx, child in enumerate(self.children)
            ]
        }
        return tree

    def _required_dependencies(self) -> Set[str]:
        deps: Set[str] = self._dependencies

        def dep_finder(item):
            nonlocal deps
            for child in item.children:
                deps = deps | set(getattr(child, "_dependencies", None))
                if hasattr(child, "children"):
                    dep_finder(child)

        dep_finder(self)
        return deps

    def _tree(self) -> str:
        return pformat(self._recurse_children(idx=0))

    def _ipython_key_completions_(self):  # pragma: no cover
        return [
            getattr(child, "title")
            for child in self.children
            if hasattr(child, "title")
        ]


class Page(Layout):
    """Defines the top level of a document.

    Args:
        title (str): Used as a title within the document and as a key value.
        navbrand (str): Brand name. Displayed in the page navbar if provided.
        children (list): Child items defining layout and content.

    """

    def __init__(
        self,
        title: Optional[str] = None,
        navbrand: Optional[str] = "",
        children: Union[
            List[Union["Layout", "Content", Any]], "Layout", "Content"
        ] = list(),
    ):
        super().__init__(title, children)
        self.org_name = navbrand

    def save(
        self,
        filepath: str = "./esparto-doc.html",
        return_html: bool = False,
        dependency_source="esparto.options",
    ) -> Optional[str]:
        """
        Save document to HTML file.

        Note: Alias for `self.save_html()`.

        Args:
          filepath (str): Destination filepath.
          return_html (bool): If True, return HTML as a string.
          dependency_source (str): One of 'cdn', 'inline', or 'esparto.options'.

        Returns:
          Document rendered as HTML. (If `return_html` is True)

        """
        html = self.save_html(
            filepath=filepath,
            return_html=return_html,
            dependency_source=dependency_source,
        )

        if return_html:
            return html
        return None

    def save_html(
        self,
        filepath: str = "./esparto-doc.html",
        return_html: bool = False,
        dependency_source="esparto.options",
    ) -> Optional[str]:
        """
        Save document to HTML file.

        Args:
          filepath (str): Destination filepath.
          return_html (bool): If True, return HTML as a string.
          dependency_source (str): One of 'cdn', 'inline', or 'esparto.options'.

        Returns:
          Document rendered as HTML. (If `return_html` is True)

        """
        html = publish_html(
            self,
            filepath=filepath,
            return_html=return_html,
            dependency_source=dependency_source,
        )

        if return_html:
            return html
        return None

    def save_pdf(
        self, filepath: str = "./esparto-doc.pdf", return_html: bool = False
    ) -> Optional[str]:
        """
        Save document to PDF file.

        Note: Requires optional module `weasyprint`.

        Args:
          filepath (str): Destination filepath.
          return_html (bool): If True, return intermediate HTML representation as a string.

        Returns:
          Document rendered as HTML. (If `return_html` is True)

        """
        html = publish_pdf(self, filepath, return_html=return_html)

        if return_html:
            return html
        return None

    _title_tags = "<h1 class='display-4 my-3'>{}</h1>"
    _body_tags = "<main class='container px-2'>{}</main>"

    @property
    def _parent_class(self):
        return Page

    @property
    def _child_class(self):
        return Section


class Section(Layout):
    """Sections define thematically distinct groups of content within a Page.

    Args:
        title (str): Used as a title within the document and as a key value.
        children (list): Child items defining layout and content.

    """

    _title_tags = "<h3 class='mb-3'>{}</h3>"
    _body_tags = "<div class='px-1 mb-5'>{}</div>"
    _parent_class = Page

    @property
    def _child_class(self):
        return Row


class Row(Layout):
    """Rows are used in combination with Columns to define the grid layout within a section.

    Args:
        title (str): Used as a title within the document and as a key value.
        children (list): Child items defining layout and content.

    """

    _title_tags = "<div class='col-12'><h5 class='px-1 mb-3'>{}</h5></div>"
    _body_tags = "<div class='row'>{}</div>"
    _parent_class = Section

    @property
    def _child_class(self):
        return Column


class Column(Layout):
    """Columns sit within Rows and act as content holders.

    Args:
        title (str): Used as a title within the document and as a key value.
        children (list): Child items defining layout and content.

    """

    _title_tags = "<h5 class='px-1 mb-3'>{}</h5>"
    _body_tags = "<div class='col-lg mb-3'>{}</div>"
    _parent_class = Row

    @property
    def _child_class(self):
        raise NotImplementedError



"""Content dependency management."""

from collections import UserDict
from dataclasses import dataclass, field
from pathlib import Path
from typing import List, Set

from esparto import _INSTALLED_MODULES, _MODULE_PATH
from esparto._options import get_source_from_options


@dataclass
class ContentDependency:
    name: str
    cdn: str
    inline: str
    target: str


@dataclass
class ResolvedDeps:
    head: List[str] = field(default_factory=list)
    tail: List[str] = field(default_factory=list)


class ContentDependencyDict(UserDict):
    def __add__(self, item: ContentDependency):
        super().__setitem__(item.name, item)
        return self


JS_DEPS = {"bokeh", "plotly"}


def lazy_content_dependency_dict() -> ContentDependencyDict:
    bootstrap_cdn = (
        '<link rel="stylesheet" href="https://stackpath.bootstrapcdn.com/bootstrap/4.4.1/css/bootstrap.min.css" '
        + 'integrity="sha384-Vkoo8x4CGsO3+Hhxv8T/Q5PaXtkKtu6ug5TOeNV6gBiFeWPGFN9MuhOf23Q9Ifjh" crossorigin="anonymous">'
    )
    bootstrap_inline = Path(_MODULE_PATH, "resources/css/bootstrap.min.css").read_text()
    bootstrap_inline = f"<style>\n{bootstrap_inline}\n</style>"

    content_dependency_dict = ContentDependencyDict()
    content_dependency_dict += ContentDependency(
        "bootstrap", bootstrap_cdn, bootstrap_inline, "head"
    )

    if "bokeh" in _INSTALLED_MODULES:
        import bokeh.resources as bk_resources  # type: ignore

        bokeh_cdn = bk_resources.CDN.render_js()
        bokeh_inline = bk_resources.INLINE.render_js()

        content_dependency_dict += ContentDependency(
            "bokeh", bokeh_cdn, bokeh_inline, "tail"
        )

    if "plotly" in _INSTALLED_MODULES:
        from plotly import offline as plotly_offline  # type: ignore

        plotly_version = "latest"
        plotly_cdn = f"<script src='https://cdn.plot.ly/plotly-{plotly_version}.min.js'></script>"
        plotly_inline = plotly_offline.get_plotlyjs()
        plotly_inline = f"<script>\n{plotly_inline}\n</script>"

        content_dependency_dict += ContentDependency(
            "plotly", plotly_cdn, plotly_inline, "head"
        )

    return content_dependency_dict


def resolve_deps(
    required_deps: Set[str], source: str = "esparto.options"
) -> ResolvedDeps:
    resolved_deps = ResolvedDeps()

    source = get_source_from_options(source)

    for dep in required_deps:
        dep_details: ContentDependency = lazy_content_dependency_dict()[dep]
        getattr(resolved_deps, dep_details.target).append(getattr(dep_details, source))

    return resolved_deps



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['esparto']

package_data = \
{'': ['*'], 'esparto': ['resources/css/*', 'resources/jinja/*']}

install_requires = \
['Pillow>=7.0.0,<9', 'jinja2>=2.10.1,<4.0.0', 'markdown>=3.1,<4.0']

extras_require = \
{':python_version < "3.7"': ['dataclasses'],
 'extras': ['beautifulsoup4>=4.7', 'weasyprint>=51']}

setup_kwargs = {
    'name': 'esparto',
    'version': '1.0.1',
    'description': 'Simple HTML and PDF document generator for Python.',
    'long_description': 'esparto\n=======\n\n[![image](https://img.shields.io/pypi/v/esparto.svg)](https://pypi.python.org/pypi/esparto)\n[![PyPI pyversions](https://img.shields.io/pypi/pyversions/esparto.svg)](https://pypi.python.org/pypi/esparto/)\n[![Build Status](https://travis-ci.com/domvwt/esparto.svg?branch=main)](https://travis-ci.com/domvwt/esparto)\n[![codecov](https://codecov.io/gh/domvwt/esparto/branch/main/graph/badge.svg?token=35J8NZCUYC)](https://codecov.io/gh/domvwt/esparto)\n[![Quality Gate Status](https://sonarcloud.io/api/project_badges/measure?project=domvwt_esparto&metric=alert_status)](https://sonarcloud.io/dashboard?id=domvwt_esparto)\n\n## Introduction\n\n`esparto` is a simple HTML and PDF document generator for Python. The API design emphasises\nproductivity and reliability over flexibility or complexity - although you should find that it serves many\nuse cases more than adequately. `esparto` is suitable for tasks such as:\n\n* Designing simple web pages\n* Automated MI reporting\n* Collating and sharing data graphics\n* ML model performance and evaluation documents\n\n\n## Main Features\n* Lightweight API\n* Jupyter Notebook support\n* Output self-contained HTML and PDF files\n* Responsive layout from [Bootstrap](https://getbootstrap.com/)\n* No CSS or HTML required\n* Automatic conversion for:\n    * Markdown\n    * Images\n    * Pandas DataFrames\n    * Matplotlib\n    * Bokeh\n    * Plotly\n\n\n## Installation\n`esparto` is available from PyPI:\n```bash\npip install esparto\n```\n\nIf PDF output is required, `weasyprint` must also be installed:\n```bash\npip install weasyprint\n```\n\n\n## Dependencies\n\n*   [python](https://python.org/) >= 3.6\n*   [jinja2](https://palletsprojects.com/p/jinja/)\n*   [markdown](https://python-markdown.github.io/)\n*   [Pillow](https://python-pillow.org/)\n*   [weasyprint](https://weasyprint.org/) _(optional - for PDF output)_\n\n\n## License\n[MIT](https://opensource.org/licenses/MIT)\n\n\n## Documentation\nFull documentation and examples are available at [domvwt.github.io/esparto/](https://domvwt.github.io/esparto/).\n\n\n## Basic Usage\n```python\nimport esparto as es\n\n# Instantiating a Page\npage = es.Page(title="Research")\n\n# Page layout hierarchy:\n# Page -> Section -> Row -> Column -> Content\n\n# Add or update content\n# Keys are used as titles\npage["Introduction"]["Part One"]["Item A"] = "lorem ipsum"\npage["Introduction"]["Part One"]["Item B"] = "./pictures/image1.jpg"\n\n# Add content without a title\npage["Introduction"]["Part One"][""] = "Hello, Wolrd!"\n\n# Replace child at index - useful if no title given\npage["Introduction"]["Part One"][-1] = "Hello, World!"\n\n# Set content and return input object\n# Useful in Jupyter Notebook as it will be displayed in cell output\npage["Methodology"]["Part One"]["Item A"] << "dolor sit amet"\n# >>> "dolor sit amet"\n\n# Set content and return new layout\npage["Methodology"]["Part Two"]["Item B"] >> "foobar"\n# >>> {\'Item B\': [\'Markdown\']}\n\n# Show document structure\npage.tree()\n# >>> {\'Research\': [{\'Introduction\': [{\'Part One\': [{\'Item A\': [\'Markdown\']},\n#                                                   {\'Item B\': [\'Image\']}]}]},\n#                   {\'Methodology\': [{\'Part One\': [{\'Item A\': [\'Markdown\']}]},\n#                                    {\'Part Two\': [{\'Item A\': [\'Markdown\']}]}]}]}\n\n# Remove content\ndel page["Methodology"]["Part One"]["Item A"]\ndel page.methodology.part_two.item_b\n\n# Access existing content as an attribute\npage.introduction.part_one.item_a = "./pictures/image2.jpg"\npage.introduction.part_one.tree()\n# >>> {\'Part One\': [{\'Item A\': [\'Image\']},\n#                   {\'Item B\': [\'Image\']},\n#                   {\'Column 2\': [\'Markdown\']}]}\n\n# Save the document\npage.save_html("my-page.html")\npage.save_pdf("my-page.pdf")\n```\n\n\n## Example Output\n\nIris Report - [HTML](https://domvwt.github.io/esparto/examples/iris-report.html) |\n[PDF](https://domvwt.github.io/esparto/examples/iris-report.pdf)\n\nBokeh and Plotly - [HTML](https://domvwt.github.io/esparto/examples/interactive-plots.html) |\n[PDF](https://domvwt.github.io/esparto/examples/interactive-plots.pdf)\n\n<br>\n\n<img width=600  src="https://github.com/domvwt/esparto/blob/fdc0e787c0bc013d16667773e82e21c647b71d91/docs/images/iris-report-compressed.png?raw=true"\nalt="example page" style="border-radius:0.5%;">\n',
    'author': 'Dominic Thorn',
    'author_email': 'dominic.thorn@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://domvwt.github.io/esparto',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6.1,<4.0',
}


setup(**setup_kwargs)
